import math
from .game import Game


class Evaluate:

    def __init__(self, base):
        self.ratings_by_players = {}
        for name, player in base.players.items():
            ratings = list(map(lambda d: [d.day, d.elo()], player.days))
            self.ratings_by_players[name] = sorted(ratings)

    def get_rating(self, name, day, ignore_null_players=True):
        min_day, min_rating = None, None
        max_day, max_rating = None, None
        if not name in self.ratings_by_players.keys():
            if ignore_null_players:
                return None
            else:
                return 0.
        ratings = self.ratings_by_players[name]
        for i in range(len(ratings)):
            if ratings[i][0] <= day:
                if (not min_day) or (ratings[i][0] >= min_day):
                    min_day = ratings[i][0]
                    min_rating = ratings[i][1]
            if ratings[i][0] >= day:
                if (not max_day) or (ratings[i][0] <= max_day):
                    max_day = ratings[i][0]
                    max_rating = ratings[i][1]
        if not min_day:
            ret = max_rating
        elif not max_day:
            ret = min_rating
        elif max_day <= min_day:
            ret = max_rating
        else:
            ret = ((max_day - day) * min_rating +
                   (day - min_day) * max_rating) * 1.0 / (max_day - min_day)
        return ret

    def evaluate_single_game(self, game, ignore_null_players=True):
        black_rating = self.get_rating(game.black_player,
                                       game.day,
                                       ignore_null_players=ignore_null_players)
        white_rating = self.get_rating(game.white_player,
                                       game.day,
                                       ignore_null_players=ignore_null_players)
        if black_rating == None or white_rating == None:
            return None
        if game.handicap_proc:
            black_advantage = game.handicap_proc(game)
        else:
            black_advantage = game.handicap
        white_gamma = 10.**(white_rating / 400.)
        black_adjusted_gamma = 10.**((black_rating + black_advantage) / 400.)

        if game.winner == 'W':
            return white_gamma / (white_gamma + black_adjusted_gamma)
        elif game.winner == 'B':
            return black_adjusted_gamma / (white_gamma + black_adjusted_gamma)
        else:
            return (white_gamma * black_adjusted_gamma)**0.5 / (
                white_gamma + black_adjusted_gamma)

    def evaluate_ave_log_likelihood_games(self,
                                          games,
                                          ignore_null_players=True):
        sum_ = 0.
        games = self.list_to_games(games)
        game_count = 0
        for game in games:
            game_likelihood = self.evaluate_single_game(
                game, ignore_null_players=ignore_null_players)
            if game_likelihood != None:
                sum_ += math.log(game_likelihood)
                game_count += 1
        return sum_ / game_count

    def list_to_games(self, game_list):
        games = []
        for game in game_list:
            black, white, winner, time_step = game[:4]
            handicap = 0
            if len(game) >= 5:
                handicap = game[4]
            extras = {}
            if len(game) >= 5:
                extras = game[5]
            games.append(
                Game(black, white, winner, time_step, handicap, extras))
        return games



import math


class UnstableRatingException(RuntimeError):
    pass


class Game:

    def __init__(self, black, white, winner, time_step, handicap=0, extras={}):
        self.day = time_step
        self.white_player = white
        self.black_player = black
        self.winner = winner
        self.extras = extras
        self.handicap = handicap if handicap else 0
        self.wpd = None
        self.bpd = None
        if callable(handicap):
            self.handicap_proc = handicap
        else:
            self.handicap_proc = None

    def opponents_adjusted_gamma(self, player):
        if self.handicap_proc:
            black_advantage = self.handicap_proc(self)
        else:
            black_advantage = self.handicap

        if player == self.white_player:
            opponent_elo = self.bpd.elo() + black_advantage
        elif player == self.black_player:
            opponent_elo = self.wpd.elo() - black_advantage
        else:
            raise ValueError(
                "No opponent for %s, since they're not in this game: %s." %
                (player.inspect(), self.inspect()))

        try:
            rval = 10.**(opponent_elo / 400.0)
        except OverflowError:
            rval = float('inf')
        if rval == 0. or math.isinf(rval) or math.isnan(rval):
            raise UnstableRatingException("bad adjusted gamma: %s" %
                                          self.inspect())

        return rval

    def opponent(self, player):
        if player == self.white_player:
            return self.black_player
        elif player == self.black_player:
            return self.white_player

    def inspect(self):
        return "%s: W:%s(%.2f) B:%s(%.2f) winner = %s, handicap = %s" % (
            self, self.white_player.name, self.wpd.r if self.wpd else '?',
            self.black_player.name, self.bpd.r if self.bpd else '?',
            self.winner, self.handicap)

    def likelihood(self):
        if self.winner == 'W':
            return self.white_win_probability()
        elif self.winner == 'B':
            return self.black_win_probability()
        else:
            return (self.white_win_probability() *
                    self.black_win_probability())**0.5

    # This is the Bradley-Terry Model
    def white_win_probability(self):
        return self.wpd.gamma() / (
            self.wpd.gamma() +
            self.opponents_adjusted_gamma(self.white_player))

    def black_win_probability(self):
        return self.bpd.gamma() / (
            self.bpd.gamma() +
            self.opponents_adjusted_gamma(self.black_player))



import math
from .player_day import PlayerDay
from .game import UnstableRatingException


class Player:

    def __init__(self, name, config):
        self.name = name
        self.debug = config.get('debug', False)
        self.w2 = (math.sqrt(config['w2']) * math.log(10.) /
                   400.)**2.  # Convert from elo^2 to r^2
        self.virtual_games = config['virtual_games']
        self.days = []

    def inspect(self):
        return "%s:(%s)" % (self, self.name)

    def log_likelihood(self):
        sum_ = 0.0
        sigma2 = self.compute_sigma2()
        n = len(self.days)
        for i in range(n):
            prior = 0
            if i < n - 1:
                rd = self.days[i].r - self.days[i + 1].r
                prior += (1. /
                          (math.sqrt(2 * math.pi * sigma2[i]))) * math.exp(
                              -(rd**2) / 2. / sigma2[i])
            if i > 0:
                rd = self.days[i].r - self.days[i - 1].r
                prior += (1. /
                          (math.sqrt(2 * math.pi * sigma2[i - 1]))) * math.exp(
                              -(rd**2) / 2. / sigma2[i - 1])
            if prior == 0:
                sum_ += self.days[i].log_likelihood()
            else:
                if math.isinf(self.days[i].log_likelihood()) or math.isinf(
                        math.log(prior)):
                    print("Infinity at %s: %f + %f: prior = %f, days = %s" %
                          (self.inspect(), self.days[i].log_likelihood(),
                           math.log(prior), prior,
                           map(lambda d: d.inspect(), self.days)))
                    exit(0)
                sum_ += self.days[i].log_likelihood() + math.log(prior)
        return sum_

    def hessian(self, days, sigma2):
        n = len(days)
        mat = [[0.0 for i in range(n)] for j in range(n)]
        for row in range(n):
            for col in range(n):
                if row == col:
                    prior = 0.
                    if row < n - 1:
                        prior += -1.0 / sigma2[row]
                    if row > 0:
                        prior += -1.0 / sigma2[row - 1]
                    mat[row][col] = days[row].log_likelihood_second_derivative(
                    ) + prior - 0.001
                elif row == col - 1:
                    mat[row][col] = 1.0 / sigma2[row]
                elif row == col + 1:
                    mat[row][col] = 1.0 / sigma2[col]
        return mat

    def gradient(self, r, days, sigma2):
        n = len(days)
        g = [0.0 for i in range(n)]
        for idx in range(n):
            day = days[idx]
            prior = 0
            if idx < n - 1:
                prior += -(r[idx] - r[idx + 1]) / sigma2[idx]
            if idx > 0:
                prior += -(r[idx] - r[idx - 1]) / sigma2[idx - 1]
            if self.debug:
                print("g[%d] = %f + %f" %
                      (idx, day.log_likelihood_derivative(), prior))
            g[idx] = day.log_likelihood_derivative() + prior
        return g

    def run_one_newton_iteration(self):
        for day in self.days:
            day.clear_game_terms_cache()

        if len(self.days) == 1:
            self.days[0].update_by_1d_newtons_method()
        elif len(self.days) > 1:
            self.update_by_ndim_newton()

    def compute_sigma2(self):
        sigma2 = [0.0 for i in range(len(self.days) - 1)]
        for i in range(len(self.days) - 1):
            d1 = self.days[i]
            d2 = self.days[i + 1]
            sigma2[i] = abs(d2.day - d1.day) * self.w2
        return sigma2

    def update_by_ndim_newton(self):
        r = list(map(lambda d: d.r, self.days))

        if self.debug:
            print("Updating %s" % self.inspect())
            for day in self.days:
                print("day[%d] r = %f" % (day.day, day.r))
                print("day[%d] win terms = %s" %
                      (day.day, day.won_game_terms()))
                print("day[%d] win games = %s" % (day.day, day.won_games))
                print("day[%d] draw terms = %s" %
                      (day.day, day.draw_game_terms()))
                print("day[%d] draw games = %s" % (day.day, day.draw_games))
                print("day[%d] lost terms = %s" %
                      (day.day, day.lost_game_terms()))
                print("day[%d] lost games = %s" % (day.day, day.lost_games))
                print("day[%d] log(p) = %f" % (day.day, day.log_likelihood()))
                print("day[%d] dlp = %f" %
                      (day.day, day.log_likelihood_derivative()))
                print("day[%d] dlp2 = %f" %
                      (day.day, day.log_likelihood_second_derivative()))

        # sigma squared (used in the prior)
        sigma2 = self.compute_sigma2()

        h = self.hessian(self.days, sigma2)
        g = self.gradient(r, self.days, sigma2)

        n = len(r)

        a = [0.0 for i in range(n)]
        d = [h[0][0]] + [0.0 for i in range(n - 1)]
        b = [h[0][1]] + [0.0 for i in range(n - 1)]

        for i in range(1, n):
            a[i] = h[i][i - 1] / d[i - 1]
            d[i] = h[i][i] - a[i] * b[i - 1]
            if i < n - 1:
                b[i] = h[i][i + 1]

        y = [g[0]] + [0.0 for i in range(n - 1)]
        for i in range(1, n):
            y[i] = g[i] - a[i] * y[i - 1]

        x = [0.0 for i in range(n)]
        x[n - 1] = y[n - 1] / d[n - 1]
        for i in range(n - 2, -1, -1):
            x[i] = (y[i] - b[i] * x[i + 1]) / d[i]

        new_r = [0.0 for i in range(n)]
        for i in range(n):
            new_r[i] = r[i] - x[i]

        for r_ in new_r:
            if r_ > 650:
                raise UnstableRatingException("Unstable r (%s) on player %s" %
                                              (new_r, self.inspect()))

        if self.debug:
            print("Hessian = %s" % h)
            print("gradient = %s" % g)
            print("a = %s" % a)
            print("d = %s" % d)
            print("b = %s" % b)
            print("y = %s" % y)
            print("x = %s" % x)
            print("%s (%s) => (%s)" % (self.inspect(), r, new_r))

        for idx in range(len(self.days)):
            day = self.days[idx]
            day.r = day.r - x[idx]

    def covariance(self):
        r = list(map(lambda d: d.r, self.days))

        sigma2 = self.compute_sigma2()
        h = self.hessian(self.days, sigma2)

        n = len(r)

        a = [0.0 for i in range(n)]
        d = [h[0][0]] + [0.0 for i in range(n - 1)]
        if n > 1:
            b = [h[0][1]] + [0.0 for i in range(n - 1)]
        else:
            b = [0.0]

        for i in range(1, n):
            a[i] = h[i][i - 1] / d[i - 1]
            d[i] = h[i][i] - a[i] * b[i - 1]
            if i < n - 1:
                b[i] = h[i][i + 1]

        dp = [0.0 for i in range(n)]
        dp[n - 1] = h[n - 1][n - 1]
        bp = [0.0 for i in range(n)]
        bp[n - 1] = h[n - 1][n - 2]
        ap = [0.0 for i in range(n)]
        for i in range(n - 2, -1, -1):
            ap[i] = h[i][i + 1] / dp[i + 1]
            dp[i] = h[i][i] - ap[i] * bp[i + 1]
            bp[i] = h[i][i - 1]

        v = [0.0 for i in range(n)]
        for i in range(n - 1):
            v[i] = dp[i + 1] / (b[i] * bp[i + 1] - d[i] * dp[i + 1])
        v[n - 1] = -1. / d[n - 1]

        mat = [[0.0 for i in range(n)] for j in range(n)]
        for row in range(n):
            for col in range(n):
                if row == col:
                    mat[row][col] = v[row]
                elif row == col - 1:
                    mat[row][col] = -a[col] * v[col]
        return mat

    def update_uncertainty(self):
        if len(self.days) > 0:
            c = self.covariance()
            u = [c[i][i] for i in range(len(self.days))]  # u = variance
            for i in range(len(self.days)):
                self.days[i].uncertainty = u[i]

    def add_game(self, game):
        if len(self.days) == 0 or self.days[-1].day != game.day:
            new_pday = PlayerDay(self, game.day)
            if len(self.days) == 0:
                new_pday.is_first_day = True
                new_pday.set_gamma(1.)
            else:
                new_pday.set_gamma(self.days[-1].gamma())
            self.days.append(new_pday)
        if game.white_player == self:
            game.wpd = self.days[-1]
        else:
            game.bpd = self.days[-1]
        self.days[-1].add_game(game)



import math


class PlayerDay:

    def __init__(self, player, day):
        self.day = day
        self.player = player
        self.is_first_day = False
        self.virtual_games = player.virtual_games
        self.won_games = []
        self.draw_games = []
        self.lost_games = []
        self.won_game_terms_cache = None
        self.draw_game_terms_cache = None
        self.lost_game_terms_cache = None

    def set_gamma(self, gamma):
        self.r = math.log(gamma)

    def gamma(self):
        return math.exp(self.r)

    def set_elo(self, elo):
        self.r = elo * (math.log(10.) / 400.0)

    def elo(self):
        return (self.r * 400.0) / (math.log(10.))

    def clear_game_terms_cache(self):
        self.won_game_terms_cache = None
        self.draw_game_terms_cache = None
        self.lost_game_terms_cache = None

    def won_game_terms(self):
        if self.won_game_terms_cache == None:
            self.won_game_terms_cache = []
            for g in self.won_games:
                other_gamma = g.opponents_adjusted_gamma(self.player)
                if other_gamma == 0. or math.isnan(other_gamma) or math.isinf(
                        other_gamma):
                    print("other_gamma (%s) = %s" %
                          (g.opponent(self.player).inspect(), other_gamma))
                self.won_game_terms_cache.append([1.0, 0.0, 1.0, other_gamma])
        return self.won_game_terms_cache

    def draw_game_terms(self):
        if self.draw_game_terms_cache == None:
            self.draw_game_terms_cache = []
            for g in self.draw_games:
                other_gamma = g.opponents_adjusted_gamma(self.player)
                if other_gamma == 0. or math.isnan(other_gamma) or math.isinf(
                        other_gamma):
                    print("other_gamma (%s) = %s" %
                          (g.opponent(self.player).inspect(), other_gamma))
                self.draw_game_terms_cache.append(
                    [0.5, 0.5 * other_gamma, 1.0, other_gamma])
            if self.is_first_day:
                for _ in range(self.virtual_games):
                    self.draw_game_terms_cache.append([
                        0.5, 0.5, 1.0, 1.0
                    ])  # draw with virtual player ranked with gamma = 1.0
        return self.draw_game_terms_cache

    def lost_game_terms(self):
        if self.lost_game_terms_cache == None:
            self.lost_game_terms_cache = []
            for g in self.lost_games:
                other_gamma = g.opponents_adjusted_gamma(self.player)
                if other_gamma == 0. or math.isnan(other_gamma) or math.isinf(
                        other_gamma):
                    print("other_gamma (%s) = %s" %
                          (g.opponent(self.player).inspect(), other_gamma))
                self.lost_game_terms_cache.append(
                    [0.0, other_gamma, 1.0, other_gamma])
        return self.lost_game_terms_cache

    def log_likelihood_second_derivative(self):
        sum_ = 0.0
        gamma = self.gamma()
        for _, _, c, d in (self.won_game_terms() + self.draw_game_terms() +
                           self.lost_game_terms()):
            sum_ += (c * d) / ((c * gamma + d)**2.0)
        if math.isnan(gamma) or math.isnan(sum_):
            print("won_game_terms = %s" % self.won_game_terms())
            print("draw_game_terms = %s" % self.draw_game_terms())
            print("lost_game_terms = %s" % self.lost_game_terms())
        return -1. * gamma * sum_

    def log_likelihood_derivative(self):
        tally = 0.0
        gamma = self.gamma()
        for _, _, c, d in (self.won_game_terms() + self.draw_game_terms() +
                           self.lost_game_terms()):
            tally += c / (c * gamma + d)
        return len(self.won_game_terms()) + 0.5 * len(
            self.draw_game_terms()) - gamma * tally

    def log_likelihood(self):
        tally = 0.0
        for a, b, c, d in self.won_game_terms():
            tally += math.log(a * self.gamma())
            tally -= math.log(c * self.gamma() + d)
        for a, b, c, d in self.draw_game_terms():
            tally += math.log(a * 2. * self.gamma()) * 0.5
            tally += math.log(b * 2.) * 0.5
            tally -= math.log(c * self.gamma() + d)
        for a, b, c, d in self.lost_game_terms():
            tally += math.log(b)
            tally -= math.log(c * self.gamma() + d)
        return tally

    def add_game(self, game):
        if (game.winner == "D"):
            self.draw_games.append(game)
        elif (game.winner == "W" and game.white_player == self.player) or (
                game.winner == "B" and game.black_player == self.player):
            self.won_games.append(game)
        else:
            self.lost_games.append(game)

    def update_by_1d_newtons_method(self):
        dlogp = self.log_likelihood_derivative()
        d2logp = self.log_likelihood_second_derivative()
        dr = dlogp / d2logp
        new_r = self.r - dr
        self.r = new_r



from distutils.core import setup

setup(
    name='whr',
    packages=['whr'],
    version='0.0.2',
    license='MIT',
    description=
    'A Python implementation of the Whole History Rating algorithm proposed by Rémi Coulom. '
    'The implementation is based on the Ruby code of GoShrine.',
    long_description=open('README.md', 'rt').read(),
    author='Tianyi Hao',
    author_email='haotianyi0@126.com',
    url='https://github.com/wind23/whole_history_rating',
    download_url=
    'https://github.com/wind23/whole_history_rating/archive/0.0.2.tar.gz',
    keywords=['WHR', 'whole history rating', 'Elo rating'],
    install_requires=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)


from .game import Game
from .player import Player
from .game import UnstableRatingException


class Base:

    def __init__(self, config={}):
        self.config = config
        self.config.setdefault('w2', 300.0)  # elo^2
        self.config.setdefault('virtual_games', 2)  # elo^2
        self.games = []
        self.players = {}

    def print_ordered_ratings(self):
        players = list(filter(lambda p: len(p.days) > 0,
                              self.players.values()))
        for p in sorted(players,
                        key=lambda p: p.days[-1].gamma(),
                        reverse=True):
            if len(p.days) > 0:
                print("%s\t%s" % (p.name, ';'.join(
                    map(lambda pd: '%d,%.2f' % (pd.day, pd.elo()), p.days))))

    def get_ordered_ratings(self):
        players = list(filter(lambda p: len(p.days) > 0,
                              self.players.values()))
        ratings = []
        for p in sorted(players,
                        key=lambda p: p.days[-1].gamma(),
                        reverse=True):
            if len(p.days) > 0:
                ratings.append((p.name, self.ratings_for_player(p.name)))
        return ratings

    def log_likelihood(self):
        score = 0.0
        for p in self.players.values():
            if len(p.days) > 0:
                score += p.log_likelihood()
        return score

    def player_by_name(self, name):
        if not name in self.players.keys():
            self.players[name] = Player(name, self.config)
        return self.players[name]

    def ratings_for_player(self, name):
        player = self.player_by_name(name)
        return list(
            map(lambda d: [d.day, d.elo(), d.uncertainty * 100.], player.days))

    def setup_game(self, black, white, winner, time_step, handicap, extras={}):
        # Avoid self-played games (no info)
        if black == white:
            raise ValueError("Invalid game (black player == white player)")

        white_player = self.player_by_name(white)
        black_player = self.player_by_name(black)
        game = Game(black_player, white_player, winner, time_step, handicap,
                    extras)
        return game

    def create_games(self, games):
        games.sort(key=lambda g: g[3])
        for game in games:
            black, white, winner, time_step = game[:4]
            handicap = 0
            if len(game) >= 5:
                handicap = game[4]
            extras = {}
            if len(game) >= 5:
                extras = game[5]
            self.create_game(black, white, winner, time_step, handicap, extras)

    def create_game(self,
                    black,
                    white,
                    winner,
                    time_step,
                    handicap=0,
                    extras={}):
        game = self.setup_game(black, white, winner, time_step, handicap,
                               extras)
        return self.add_game(game)

    def add_game(self, game):
        game.white_player.add_game(game)
        game.black_player.add_game(game)
        if game.bpd == None:
            print("Bad game: %d" % game.inspect())
        self.games.append(game)
        return game

    def iterate_until_converge(self, verbose=True):
        count = 0
        last_ratings = None
        best_iteration = None
        while True:
            sorted_players = sorted(self.players.items(), key=lambda x: x[0])
            ratings = []
            for player in sorted_players:
                for day in player[1].days:
                    ratings.append(round(day.elo() * 100))
            delta = 0
            if count > 0:
                for i in range(len(ratings)):
                    delta += abs(ratings[i] - last_ratings[i])
                if verbose:
                    print('Iteration: %d, delta: %d' % (count, delta))
                if delta > 0:
                    best_iteration = count
                if count - best_iteration >= 10:
                    break
                last_ratings = ratings
            else:
                best_iteration = count
                last_ratings = ratings
            self.run_one_iteration()
            count += 1
        for player in self.players.values():
            player.update_uncertainty()
        return count

    def iterate(self, count):
        for _ in range(count):
            self.run_one_iteration()
        for player in self.players.values():
            player.update_uncertainty()

    def run_one_iteration(self):
        for player in self.players.values():
            player.run_one_newton_iteration()
